/**
 * Copyright (c) 2016 - 2019, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include "nordic_common.h"
#include "app_error.h"
#include "app_uart.h"
#include "ble_db_discovery.h"
#include "app_timer.h"
#include "app_util.h"
#include "bsp_btn_ble.h"
#include "ble.h"
#include "ble_gap.h"
#include "ble_hci.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"
#include "ble_nus_c.h"
#include "nrf_ble_gatt.h"
#include "nrf_pwr_mgmt.h"
#include "nrf_ble_scan.h"

#include "nrf_drv_clock.h"
#include "app_scheduler.h"

#include "SEGGER_RTT.h"

// GUID gen for xTag online - 10c66c52-e5c9-4fe7-8a04-e40b13eb932c
#define BLE_UUID_XTAG_BASE_UUID             { 0x10, 0xC6, 0x6C, 0x52, 0xE5, 0xC9, 0x4F, 0xE7, 0x8A, 0x04, 0xE4, 0x0B, 0x13, 0xEB, 0x93, 0x2C }
#define BLE_UUID_XTAG_SERVICE               0x0001 
#define BLE_UUID_XTAG_RX_CHARACTERISTC_UUID 0x0002
#define BLE_UUID_XTAG_TX_CHARACTERISTC_UUID 0x0003

// #define BLE_UUID_XTAG_SERVICE   0x0001                         // Must match the service UUID in the xtag.
#define XTAG_SERVICE_UUID_TYPE  BLE_UUID_TYPE_VENDOR_BEGIN        // UUID type for the xtag svc.

#define APP_BLE_CONN_CFG_TAG    1                                 // Tag that refers to the BLE stack configuration set with @ref sd_ble_cfg_set. The default tag is @ref BLE_CONN_CFG_TAG_DEFAULT. */
#define APP_BLE_OBSERVER_PRIO   3                                 // BLE observer priority of the application. There is no need to modify this value. */

#define SCAN_INTERVAL           128                               // Scan interval in 625 us units (240 = 150 mSec, 64 = 40 mSec).
#define SCAN_WINDOW             64                              // Scan window in 625 us units (112 = 70 mSec, 64 = 40 mSec).
//#define SCAN_DURATION           600                             // Set explicitly via api.

#define MIN_CONN_INTERVAL       MSEC_TO_UNITS(80,UNIT_1_25_MS)   // Min acceptable con int. Def is 7.5 mSec. Set peripherals to match.
#define MAX_CONN_INTERVAL       MSEC_TO_UNITS(80, UNIT_1_25_MS)   // Max acceptable con int. Def is 30 mSec. Set peripherals to match.
#define SLAVE_LATENCY           0                                 // Match egs.
#define CONN_SUP_TIMEOUT        MSEC_TO_UNITS(4000, UNIT_10_MS)   // Match egs.

// Scheduler settings
#define SCHED_MAX_EVENT_DATA_SIZE   256  // Larger needed for our own stuff.
#define SCHED_QUEUE_SIZE            32

NRF_BLE_GATT_DEF(m_gatt);                                               /**< GATT module instance. */
NRF_BLE_SCAN_DEF(m_scan);                                               /**< Scanning Module instance. */

APP_TIMER_DEF(m_10_Sec_timer_id);                   // timer.

// Add to array within xgateway_data.c
static uint16_t m_rx_char_value_handle = BLE_CONN_HANDLE_INVALID; // Charactoristic VALUE declaration handle.
static uint16_t m_tx_char_value_handle = BLE_CONN_HANDLE_INVALID; // Charactoristic VALUE declaration handle.
static uint16_t m_tx_char_descript_handle = BLE_CONN_HANDLE_INVALID; // Charactoristic DESCRIPTOR declaration handle.
static uint16_t m_tx_handle = BLE_CONN_HANDLE_INVALID;

static ble_gattc_handle_range_t m_char_disco_handle_range;

static uint8_t m_ble_rx_buffer[256];  // Data from BLE peers. Max 244 bytes.
static uint16_t m_ble_rx_buffer_size;

// Maximum length of data (in bytes) that can be transmitted to the peer by the Nordic UART service module.
static uint16_t m_ble_max_data_len = BLE_GATT_ATT_MTU_DEFAULT - OPCODE_LENGTH - HANDLE_LENGTH; 

static ble_uuid_t const m_xtag_svc_uuid =  
{
    .uuid = BLE_UUID_XTAG_SERVICE,       // Scan only for devices with our xtag service.
    .type = XTAG_SERVICE_UUID_TYPE
};

static ble_gap_scan_params_t const m_scan_param =
{
    .active        = 0x01,
    .interval      = SCAN_INTERVAL,
    .window        = SCAN_WINDOW,
    .filter_policy = BLE_GAP_SCAN_FP_ACCEPT_ALL, //BLE_GAP_SCAN_FP_WHITELIST,
    .timeout       = BLE_GAP_SCAN_TIMEOUT_UNLIMITED,
    .scan_phys     = BLE_GAP_PHY_1MBPS,
    .channel_mask[4] = 0x00, // Use 0xC0 if you want to block channels 39 and 38 and only scan 37. 3 chan besat for noisy areas.
};

static ble_gap_conn_params_t const m_con_param =
{
  .min_conn_interval  = MIN_CONN_INTERVAL,
  .max_conn_interval  = MAX_CONN_INTERVAL,
  .slave_latency      = SLAVE_LATENCY,
  .conn_sup_timeout   = CONN_SUP_TIMEOUT,
};

static uint16_t m_notificationCount = 0;

void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(0xDEADBEEF, line_num, p_file_name);
}

static void timer_handler(void* p_context)
{
  // This handler was used initially to stop scanning in 10 sec - now we stop scanning whenever a device is found and we can connect.

  //nrf_ble_scan_stop();
  //SEGGER_RTT_printf(0, "Scanning stopped.\n");

  return;
}

void on_main_ble_sched_handler(void *p_event_data, uint16_t event_size)
{
  // Handle incoming notification data that is preset. Note that args p_event_data and size are unused.
  SEGGER_RTT_printf(0, "Recieved notifcation %d with length %d. Starting with data %d %d.\n", ++m_notificationCount, m_ble_rx_buffer_size, m_ble_rx_buffer[0], m_ble_rx_buffer[1]);
}

// Fcn that confirms scanned device / peripheral is ours.
bool ble_adv_contains_mfrid(uint8_t* pAdvADBlocks, uint8_t AdvLen, uint16_t mfrId)
{
  uint8_t curIdx = 0;
  uint8_t curADLen = 0;
  uint8_t sanityCnt = 0;
  
  // Separate ms and ls bytes in mfrId.
  uint8_t msMfrByte = (uint8_t) ((mfrId >> 8) & 0x00FF);
  uint8_t lsMfrByte = (uint8_t) ((mfrId >> 0) & 0x00FF);
  uint8_t cnt = 0;

  // Walk through Adv packet (max len 32 bytes) looking for individual AD blocks by looking at each AD block length, starting at 0.
  while(curIdx < AdvLen)
  {
    // Get the cur AD block len using the idx.
    curADLen = pAdvADBlocks[curIdx];

    // Misalign if we ever get a length of zero or beyond AdvLen.
    if((curADLen == 0) || (curIdx + (curADLen + 1)) > AdvLen) { SEGGER_RTT_printf(0, "Bad len %d. Fail out.\n", curADLen); return false; }
    
    // If mfr block with type 0xFF, check mfrId.
    if(pAdvADBlocks[curIdx + 1] == 0xFF)
    {
      // If backwards mfrId match return true, else false.
      if((pAdvADBlocks[curIdx + 2] == lsMfrByte) && (pAdvADBlocks[curIdx + 3] == msMfrByte)) { return true;}
      else { return false; }
    }
    else { curIdx += curADLen + 1; }  // Advance to the next AD block (length in packet does not include its own byte).

    // Increase the sanity count and return if it gets high as we cannot stay in here forever.
    if(++sanityCnt > 10) { return false; }
  }

  // Should never get here as we should always find an AD block with type 0xFF (mfr data).
  return false;
}

static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
  ret_code_t            err_code;
  ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;
  ble_gap_evt_data_length_update_t const* p_datalen_evt;

  switch (p_ble_evt->header.evt_id)
  {
    case BLE_GAP_EVT_ADV_REPORT:

      // Look for Dworx Mfr ID 0x030E.
      if(ble_adv_contains_mfrid(p_ble_evt->evt.gap_evt.params.adv_report.data.p_data,
                                    p_ble_evt->evt.gap_evt.params.adv_report.data.len, 0x030E))
      {
        SEGGER_RTT_printf(0, "Scanned xTAG. Will stop scanning and connect.\n");

        nrf_ble_scan_stop();

        // Connect using the address and other memebers.
        // Connect status in our store will be updated when a connected event comes in.
        err_code = sd_ble_gap_connect(&(p_ble_evt->evt.gap_evt.params.adv_report.peer_addr), &m_scan_param, &m_con_param, APP_BLE_CONN_CFG_TAG);

        // Provide status.
        if(err_code != NRF_SUCCESS) 
        {
          if(err_code == NRF_ERROR_CONN_COUNT)
          {
            SEGGER_RTT_printf(0, "Connection error. Max connections exceeded.\n"); 
          }
          else
          {
            SEGGER_RTT_printf(0, "Connection error %d\n", err_code); 
          }
        }
        else 
        { 
          SEGGER_RTT_printf(0, "Connected to xTAG.\n"); 
        }
      }

      break;

    case BLE_GAP_EVT_CONNECTED:
    {
      SEGGER_RTT_printf(0, "Connected to tag\n");

      // Request 1MBPS or 2MBPS or Coded phy on the peripheral. 
      // NOTE: The peripheral phy update request will only go out if it is different than the current connection phy.
      ble_gap_phys_t phys;
      phys.rx_phys = BLE_GAP_PHY_2MBPS;
      phys.tx_phys = BLE_GAP_PHY_2MBPS;
      err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
      APP_ERROR_CHECK(err_code);

      // Log the effective mtu.
      SEGGER_RTT_printf(0, "Effective MTU after connection is: %d bytes\n", 
        nrf_ble_gatt_eff_mtu_get(&m_gatt, p_ble_evt->evt.gap_evt.conn_handle));

      break;
    }
    case BLE_GAP_EVT_DISCONNECTED:
    {
      uint8_t Idx;

      if(p_gap_evt->params.disconnected.reason == BLE_HCI_CONN_FAILED_TO_BE_ESTABLISHED)
      {
        SEGGER_RTT_printf(0, "Connection Failed!\n", err_code);
      }
      else
      {
        SEGGER_RTT_printf(0, "Unsolicited disconnection from tag with handle %04X\n", 
         p_ble_evt->evt.gap_evt.conn_handle);
      }

      break;
    }
    case BLE_GAP_EVT_TIMEOUT:

      //SEGGER_RTT_printf(0, "BLE_GAP_EVT_TIMEOUT\n.");

      if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
      {
        SEGGER_RTT_printf(0, "Connection Request timed out.\n");
      }
      else if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_SCAN)
      {
        SEGGER_RTT_printf(0, "Scan timed out.\n");
      }
      break;

    case BLE_GAP_EVT_PHY_UPDATE:
      
      SEGGER_RTT_printf(0, "Phy updated. rx %d tx %d\n", 
      p_gap_evt->params.phy_update.rx_phy, p_gap_evt->params.phy_update.tx_phy);

      // Start service discovery.
      err_code = sd_ble_gattc_primary_services_discover(p_ble_evt->evt.gap_evt.conn_handle, 0x0B, &m_xtag_svc_uuid);
      APP_ERROR_CHECK(err_code);

      break;

    case BLE_GAP_EVT_SEC_PARAMS_REQUEST:

      // Pairing not supported.
      err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
      APP_ERROR_CHECK(err_code);
      break;

    case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
        
      // Accepting parameters requested by peer.
      SEGGER_RTT_printf(0, "Connection param update from peer. Min 1.25 mSec units %d. Max 1.25 mSec units %d\n", 
        p_gap_evt->params.conn_param_update_request.conn_params.min_conn_interval,
        p_gap_evt->params.conn_param_update_request.conn_params.max_conn_interval);

      err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
                                              &p_gap_evt->params.conn_param_update_request.conn_params);
      APP_ERROR_CHECK(err_code);
      break;

    case BLE_GAP_EVT_CONN_PARAM_UPDATE:

      SEGGER_RTT_printf(0, "Connection param update. Min 1.25 mSec units %d. Max 1.25 mSec units %d\n", 
        p_gap_evt->params.conn_param_update.conn_params.min_conn_interval,
        p_gap_evt->params.conn_param_update.conn_params.max_conn_interval);

        break;
      
    case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
    {
      // We don't handle peripheral requests. We make the request after connection.
      SEGGER_RTT_printf(0, "PHY update request in. Will ignore requested rx phy %d and tx phy %d.\n",
       p_ble_evt->evt.gap_evt.params.phy_update_request.peer_preferred_phys.rx_phys,
       p_ble_evt->evt.gap_evt.params.phy_update_request.peer_preferred_phys.tx_phys);

      break;
    }

    case BLE_GAP_EVT_DATA_LENGTH_UPDATE:
    {
      p_datalen_evt = &p_ble_evt->evt.gap_evt.params.data_length_update;

      SEGGER_RTT_printf(0, "Data length (PDU) update. Max Tx octets: %d, Max Rx octets %d, Max Tx uSec: %d, Max Rx uSec: %d\n",
        p_datalen_evt->effective_params.max_tx_octets, p_datalen_evt->effective_params.max_rx_octets, 
        p_datalen_evt->effective_params.max_tx_time_us, p_datalen_evt->effective_params.max_rx_time_us);

        break;
    }

    case BLE_GATTC_EVT_TIMEOUT:
        
      // Disconnect on GATT Client timeout event.
      SEGGER_RTT_printf(0, "GATT Client Timeout.\n");
      err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                       BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
      APP_ERROR_CHECK(err_code);
      break;

    case BLE_GATTS_EVT_TIMEOUT:
        
      // Disconnect on GATT Server timeout event.
      SEGGER_RTT_printf(0, "GATT Server Timeout.\n");
      err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                       BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
      APP_ERROR_CHECK(err_code);
      break;

    case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP :
    {
      ble_gattc_evt_t const * p_gatt_evt = &p_ble_evt->evt.gattc_evt;
      uint16_t num_svcs = p_gatt_evt->params.prim_srvc_disc_rsp.count;

      // Confirm status and at least 1 service was found.
      if((p_gatt_evt->gatt_status != NRF_SUCCESS) || (num_svcs < 1))
      {
        SEGGER_RTT_printf(0, "GATT Svc resp with bad status of %d or num svcs of %d.\n",
            p_gatt_evt->gatt_status, num_svcs);
        break;
      }

      // If the UUID does not match - ignore. We should only get 1 svc.
      if(p_gatt_evt->params.prim_srvc_disc_rsp.services[0].uuid.uuid != BLE_UUID_XTAG_SERVICE)
      {
        SEGGER_RTT_printf(0, "GATT Svc disco resp with bad UUID of %d\n",
            p_gatt_evt->params.prim_srvc_disc_rsp.services[0].uuid.uuid);
        break;
      }

      for(int x=0; x<num_svcs; x++) { SEGGER_RTT_printf(0, "Primary svc found with UUID %d, handle range start %d and end %d.\n", 
            p_gatt_evt->params.prim_srvc_disc_rsp.services[x].uuid.uuid,
            p_gatt_evt->params.prim_srvc_disc_rsp.services[x].handle_range.start_handle,
            p_gatt_evt->params.prim_srvc_disc_rsp.services[x].handle_range.end_handle); }

      // Setup handle ranges to look for chars on the server.
      m_char_disco_handle_range.start_handle = p_gatt_evt->params.prim_srvc_disc_rsp.services[0].handle_range.start_handle;
      m_char_disco_handle_range.end_handle = p_gatt_evt->params.prim_srvc_disc_rsp.services[0].handle_range.end_handle;

      // Clear handles. These will be setup within charactoristic (rx) and descriptor (tx) disco event handling.
      m_rx_char_value_handle = BLE_CONN_HANDLE_INVALID;
      m_tx_char_value_handle = BLE_CONN_HANDLE_INVALID;

      // Results in BLE_GATTC_EVT_CHAR_DISC_RSP event (case below).
      err_code = sd_ble_gattc_characteristics_discover(p_gatt_evt->conn_handle, &m_char_disco_handle_range);
      APP_ERROR_CHECK(err_code);

      break;
    }

    case BLE_GATTC_EVT_CHAR_DISC_RSP:
    {
      // Look through charactoristics and retrieve char declaration and char value declaration handles as req'd.
      // Critically - for each char, the UUID set will be for the char value declaration and not the char declaration
      // as the char declaration UUID is always static (0x2803).

      // Setup reusable vars.
      ble_gattc_evt_t const * p_gatt_evt = &p_ble_evt->evt.gattc_evt;
      uint16_t num_chars = p_gatt_evt->params.char_disc_rsp.count;

      // DEBUG
      SEGGER_RTT_printf(0, "GATT char disco with %d chars\n", num_chars);

      // Confirm status
      if(p_gatt_evt->gatt_status != NRF_SUCCESS)
      {
        SEGGER_RTT_printf(0, "GATT char disco resp with bad status of %d.\n", p_gatt_evt->gatt_status);

        break;
      }

      // Loop through chars looking for our Rx UUID
      ble_gattc_char_t idx_char;
      for(int x=0; x<num_chars; x++) 
      {
        // Setup indexed charactoristic.
        idx_char = p_gatt_evt->params.char_disc_rsp.chars[x];

        SEGGER_RTT_printf(0, "GATT char disco with UUID %04X and declaration handle %04X\n", idx_char.uuid.uuid, idx_char.handle_decl);

        if(idx_char.uuid.uuid == BLE_UUID_XTAG_RX_CHARACTERISTC_UUID) 
        {
          // Store the handle for subsequent writes.
          m_rx_char_value_handle = idx_char.handle_value;

          // Increment the start handle for subsequent char discoveries.
          m_char_disco_handle_range.start_handle = idx_char.handle_decl + 1;
        }
        else if(idx_char.uuid.uuid == BLE_UUID_XTAG_TX_CHARACTERISTC_UUID)
        {
          // Set the m_tx_char_value_handle to match the char value handle.
          m_tx_char_value_handle = idx_char.handle_value; 

          // We can start to look for the descriptor for this char.
          // ... handled within the BLE_GATTC_EVT_DESC_DISC_RSP event case that follows.
          // The descriptor handle should follow the char value declaration handle. Set start_handle.
          m_char_disco_handle_range.start_handle = idx_char.handle_value + 1;
          err_code = sd_ble_gattc_descriptors_discover(p_gatt_evt->conn_handle, &m_char_disco_handle_range);
          APP_ERROR_CHECK(err_code);
        }
        else
        {
          // Just inc the start_handle if ther is no match.
          m_char_disco_handle_range.start_handle = idx_char.handle_decl + 1;
        }
      }

      // If either handle is not set ...
      if( (m_rx_char_value_handle == BLE_CONN_HANDLE_INVALID) || (m_tx_char_value_handle == BLE_CONN_HANDLE_INVALID))
      {
        // If we are at the end of the handles, give up.
        if(m_char_disco_handle_range.start_handle > m_char_disco_handle_range.end_handle)
        {
          SEGGER_RTT_printf(0, "GATT char disco resp with missing UUIDs.\n");
        }
        else
        {
          // Try again using the updated handle range.
          err_code = sd_ble_gattc_characteristics_discover(p_gatt_evt->conn_handle, &m_char_disco_handle_range);
          APP_ERROR_CHECK(err_code);
        }
      }
      
      break;
    }

    case BLE_GATTC_EVT_DESC_DISC_RSP:
    {
      // Setup reusable vars.
      ble_gattc_evt_t const * p_gatt_evt = &p_ble_evt->evt.gattc_evt;
      uint16_t num_descriptors = p_gatt_evt->params.desc_disc_rsp.count;

      // DEBUG
      SEGGER_RTT_printf(0, "GATT descriptor disco with %d descriptors\n", num_descriptors);

      // Confirm status
      if(p_gatt_evt->gatt_status != NRF_SUCCESS)
      {
        SEGGER_RTT_printf(0, "GATT descriptor disco resp with bad status of %d.\n", p_gatt_evt->gatt_status);

        break;
      }

      // Loop through descriptors looking for our CCCD UUID. Because we setup this event with the handle that
      // should match our tx char descriptor, we should only get 1 event and it should match.
      ble_gattc_desc_t idx_descriptor;
      for(int x=0; x<num_descriptors; x++) 
      {
        SEGGER_RTT_printf(0, "GATT descriptor disco with UUID %04X and handle %04X\n", 
          p_gatt_evt->params.desc_disc_rsp.descs[x].uuid.uuid,
          p_gatt_evt->params.desc_disc_rsp.descs[x].handle );

        if(p_gatt_evt->params.desc_disc_rsp.descs[x].uuid.uuid == BLE_UUID_DESCRIPTOR_CLIENT_CHAR_CONFIG)
        {
          // Simply use the descriptor handle to write a value to enable notificatins on the tx char.
          // First - create write_params and and a byte array to write.
          ble_gattc_write_params_t write_params;
          memset(&write_params, 0, sizeof(write_params));
          m_tx_char_descript_handle = p_gatt_evt->params.desc_disc_rsp.descs[x].handle;

          uint8_t cccd_value[2] = { BLE_GATT_HVX_NOTIFICATION, 0 };       // 1 to enable notifications.

          write_params.handle = m_tx_char_descript_handle;
          write_params.len = 2;
          write_params.p_value = cccd_value;
          write_params.write_op = BLE_GATT_OP_WRITE_REQ;
          write_params.offset = 0;
          write_params.flags = BLE_GATT_EXEC_WRITE_FLAG_PREPARED_WRITE;

          err_code = sd_ble_gattc_write(p_gatt_evt->conn_handle, &write_params);
          APP_ERROR_CHECK(err_code);

          return; // Only need the 1 descriptor.
        }
      }

      // If we get here, there is an issue finding our descriptor.
      SEGGER_RTT_printf(0, "GATT descriptor disco fail. No cccd found\n");
      
      break;
    }

    case BLE_GATTC_EVT_WRITE_RSP:
    {
      // Setup reusable vars.
      ble_gattc_evt_t const * p_gatt_evt = &p_ble_evt->evt.gattc_evt;

      // Handle descripter var write responses.
      if(p_gatt_evt->params.write_rsp.handle == m_tx_char_descript_handle)
      {
        // Notification setup complete. If the other handles are valid, connection is complete.
        if((m_rx_char_value_handle != BLE_CONN_HANDLE_INVALID) && (m_tx_char_value_handle != BLE_CONN_HANDLE_INVALID))
        {
          // Store the handle for tx. Wont be used as we will handle notifications only.
          m_tx_handle = p_gatt_evt->conn_handle;

          // DEBUG
          SEGGER_RTT_printf(0, "Connect success with handle %04X. Sending cmd to start notifications\n", p_gatt_evt->conn_handle);

          // Send simple made up Cmd to start data notifications. 
          uint8_t cmd_len = 2;
          uint8_t cmdBuf[cmd_len];
          cmdBuf[0] = 0xAA;    
          cmdBuf[1] = 0xBB;   

          // Setup a ble_gattc_write_params_t for the write.
          ble_gattc_write_params_t write_params;
          memset(&write_params, 0, sizeof(write_params));
          write_params.handle = m_rx_char_value_handle;
          write_params.len = cmd_len;
          write_params.p_value = cmdBuf;
          write_params.write_op = BLE_GATT_OP_WRITE_CMD;
          write_params.offset = 0;

          // Actually send the cmd.
          err_code = sd_ble_gattc_write(m_tx_handle, &write_params);
          APP_ERROR_CHECK(err_code);

        }
        else
        {
          SEGGER_RTT_printf(0, "Connect fail. GATT Notification ok but error on rx or tx char handle.\n");
        }
      }

      break;
    }

    case BLE_GATTC_EVT_HVX:
    {
      // Setup reusable vars.
      ble_gattc_evt_t const * p_gatt_evt = &(p_ble_evt->evt.gattc_evt);

      // Setup data for subsequent processing on main.
      memcpy_fast(m_ble_rx_buffer, p_gatt_evt->params.hvx.data, (size_t) p_gatt_evt->params.hvx.len);
      m_ble_rx_buffer_size = p_gatt_evt->params.hvx.len;

      // Schedule an event on the main thread to process the response.
      app_sched_event_put(NULL,  0, on_main_ble_sched_handler);

      break;
    }

    default:

        break;
  }
}

/**@brief Function for handling events from the GATT library. */
void ble_gatt_evt_handler(nrf_ble_gatt_t * p_gatt, nrf_ble_gatt_evt_t const * p_evt)
{
  ret_code_t err_code;

  if (p_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED)
  {
    // Our NRF_SDH_BLE_GAP_DATA_LENGTH is setup as 251 in sdk_config.h. This is the PDU max length.
    // After connection, the central tries to update the MTU in the peripheral using NRF_SDH_BLE_GAP_DATA_LENGTH.
    // If successful, we get the ATT header (3 byte opcode and handle) and 244 byte ATT body back as 247 total or mtu effective. 
    // We can only use the ATT body, so set m_ble_max_data_len to that.
    // Good info here: https://punchthrough.com/maximizing-ble-throughput-part-3-data-length-extension-dle-2/
    m_ble_max_data_len = p_evt->params.att_mtu_effective - 3; // 1 Byte for att opcode and 2 for att handle (3 byte att header).
    SEGGER_RTT_printf(0, "BLE GATT ATT MTU update done. MTU effective in %d. Max data len set to %d\n",
      p_evt->params.att_mtu_effective, m_ble_max_data_len);
  }
  else if(p_evt->evt_id == NRF_BLE_GATT_EVT_DATA_LENGTH_UPDATED)
  {
    SEGGER_RTT_printf(0, "BLE GATT ATT data length update done. Event data len %d\n", p_evt->params.data_length);
  }
}

void ble_init(void)
{
  ret_code_t err_code;
  ble_opt_t ble_opt;

  ///////////////////////////////////////////////////////////////////
  // First - init the stack /////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////

  // Init BLE with NUS. BLE stuff in xgateway_bluetooth.*
  err_code = nrf_sdh_enable_request();
  APP_ERROR_CHECK(err_code);

  // Configure the BLE stack using the default settings.
  // Fetch the start address of the application RAM.
  uint32_t ram_start = 0;
  err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
  APP_ERROR_CHECK(err_code);

  // Enable BLE stack.
  err_code = nrf_sdh_ble_enable(&ram_start);
  APP_ERROR_CHECK(err_code);

  // Register a handler for BLE events.
  NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);

  ///////////////////////////////////////////////////////////////////
  // End stack init.
  ///////////////////////////////////////////////////////////////////

  // GAP - Add extended connection event support to send multiple packets per event.
  memset(&ble_opt, 0x00, sizeof(ble_opt));
  ble_opt.common_opt.conn_evt_ext.enable = 1;
  err_code = sd_ble_opt_set(BLE_COMMON_OPT_CONN_EVT_EXT, &ble_opt);
  APP_ERROR_CHECK(err_code);

  ///////////////////////////////////////////////////////////////////
  // GATT init            
  ///////////////////////////////////////////////////////////////////

  err_code = nrf_ble_gatt_init(&m_gatt, ble_gatt_evt_handler);
  APP_ERROR_CHECK(err_code);

  // Declare 16-bit service and 128-bit base UUIDs and add them to the BLE stack
  ble_uuid_t service_uuid;
  ble_uuid128_t base_uuid = BLE_UUID_XTAG_BASE_UUID;
  service_uuid.uuid = BLE_UUID_XTAG_SERVICE;
  err_code = sd_ble_uuid_vs_add(&base_uuid, &service_uuid.type);
  APP_ERROR_CHECK(err_code);

  ///////////////////////////////////////////////////////////////////
  // End GATT init             
  ///////////////////////////////////////////////////////////////////

  // Init scan and set params explicitly. Both always return NRF_SUCCESS
  nrf_ble_scan_init(&m_scan, NULL, NULL);
  nrf_ble_scan_params_set(&m_scan, &m_scan_param);
}

int main(void)
{
    // Initialize.
    SEGGER_RTT_Init();
    SEGGER_RTT_printf(0, "Central Logging started.\n");

    // As per egs, setup the clock driver to use the lfclk (before timers_init)
    APP_ERROR_CHECK(nrf_drv_clock_init());
    nrf_drv_clock_lfclk_request(NULL);

    APP_SCHED_INIT(SCHED_MAX_EVENT_DATA_SIZE, SCHED_QUEUE_SIZE);

    // Init timers.
    APP_ERROR_CHECK(app_timer_init());
    APP_ERROR_CHECK(app_timer_create(&m_10_Sec_timer_id, APP_TIMER_MODE_SINGLE_SHOT , timer_handler));

    APP_ERROR_CHECK(nrf_pwr_mgmt_init());

    ble_init();

    SEGGER_RTT_printf(0, "BLE central example started! Will try to scan peer in the next 10 sec ...\n");

    // Start scanning.
    nrf_ble_scan_start(&m_scan);

    // Start timeer with handler that will stop scanning.
    APP_ERROR_CHECK(app_timer_start(m_10_Sec_timer_id, APP_TIMER_TICKS(10000), NULL));

    // Enter main loop.
    for (;;) 
    { 
      app_sched_execute();

      // Manage power.
      nrf_pwr_mgmt_run(); 
    }
}
